/*
 * intr.c
 *
 * This file contains interrupt service routines (ISRs).
 * ISRs used during the booting process are placed at the
 * memorable interrupt vectors DE, AD, BE, EF, CA & FE.
 * The main ISR once the system is booted is placed at 80.
 *
 * This file belongs to PGOS.
 *
 */

#include <intr.h>
#include <panic.h>

/* Assembly wrapper labels for the C ISRs we define below */
extern uint32_t _isr0;
extern uint32_t _isr6;
extern uint32_t _isr8;
extern uint32_t _isr11;
extern uint32_t _isr13;

void isr_0 () {
  panic("Division by zero!");
}

void isr_6 () {
  panic("Undefined opcode!");
}

void isr_8 () {
  panic("Double fault!");
}

void isr_11 () {
  panic("Segment not present!");
}

void isr_13 () {
  panic("General protection fault!");
}

gdt_entry_t fill_gdt_entry (uint32_t limit, uint32_t base, uint8_t access) {
  gdt_entry_t entry;
  uint8_t granularity = 0;

  if (limit > 0x100000) {
    granularity = 1;
    limit >>= 12;
  }

  entry.lim1 = limit & 0xFFFF;
  entry.lim2_flags = (limit >> 16) | (granularity << 7) | 0x40;

  entry.base1 = base & 0xFFFF;
  entry.base2 = (base >> 16) & 0xFF;
  entry.base3 = base >> 24;

  entry.access = access;
  return entry;
}

void lgdt (gdt_entry_t *entries, size_t count) {
  gdtr_t gdtr;

  if (count*sizeof(gdt_entry_t) > 0x10000) {
    return;
  }
  
  gdtr.size = count*sizeof(gdt_entry_t)-1;
  gdtr.addr = (uint32_t) entries;

  __asm__ __volatile__ ( "lgdt %0" : : "m" (gdtr) : );
}

void load_default_gdt () {
  gdt_entry_t entries[3];

  entries[0] = fill_gdt_entry(0, 0, 0);
  entries[1] = fill_gdt_entry(0xFFFFFFFF, 0, 0x9A);
  entries[2] = fill_gdt_entry(0xFFFFFFFF, 0, 0x93);

  lgdt((gdt_entry_t*) entries, 3);
}

idt_entry_t fill_idt_entry (uint32_t offset, uint8_t flags, uint16_t segment) {
  idt_entry_t entry;

  entry.offset1 = offset & 0xFFFF;
  entry.segment = 8*segment;
  entry.resv = 0;
  entry.flags = flags;
  entry.offset2 = offset >> 16;
  
  return entry;
}

void lidt (idt_entry_t *entries) {
  gdtr_t idtr;

  idtr.size = 256*sizeof(idt_entry_t)-1;
  idtr.addr = (uint32_t) entries;

  __asm__ __volatile__ ( "lidt %0" : : "m" (idtr) : );
}

void load_default_idt () {
  idt_entry_t entries[256];
  size_t idx;

  for (idx = 0; idx < 256; idx++) {
    entries[idx] = fill_idt_entry(0,0,0);
  }

  entries[0] = fill_idt_entry(_isr0, 0x8E, 1);
  entries[6] = fill_idt_entry(_isr6, 0x8E, 1);
  entries[8] = fill_idt_entry(_isr8, 0x8E, 1);
  entries[11] = fill_idt_entry(_isr11, 0x8E, 1);
  entries[13] = fill_idt_entry(_isr13, 0x8E, 1);

  lidt((idt_entry_t*) entries);
}
