/*
 * vga.c
 *
 * This file contains a VGA text mode driver.
 * This file belongs to PGOS.
 *
 */

#include <term/vga.h>

static size_t terminal_row;
static size_t terminal_column;
static uint8_t terminal_color;
static volatile uint16_t *terminal_buffer;

/* An implementation of string.h's strlen, will later be moved to own file */
size_t strlen(const char* str) {
  size_t len = 0;
  
  while (str[len] != 0) {
    len++;
  }
  return len;
}

/* Increments the cursor position, with line / page wrap */
void inc_wrap () {
  terminal_column++;
  
  if (terminal_column == VGA_WIDTH) {
    terminal_column = 0;
    terminal_row = (terminal_row + 1 < VGA_HEIGHT) ? terminal_row + 1 : 0;
  }
}

/* Decrements the cursor position, with line / page wrap */
void dec_wrap () {
  terminal_column = (terminal_column > 0) ? terminal_column - 1 : VGA_WIDTH - 1;

  if (terminal_column == VGA_WIDTH - 1) {
    terminal_row = (terminal_row > 0) ? terminal_row - 1 : 0;
  }
}

/* Sets all 2000 cells to whitespace */
void terminal_clear () {
  size_t index;
  size_t y;
  size_t x;

  for (y = 0; y < VGA_HEIGHT; y++) {
    for (x = 0; x < VGA_WIDTH; x++) {
      index = y * VGA_WIDTH + x;
      terminal_buffer[index] = vga_entry(' ', terminal_color);
    }
  }
}

/* Carriage return & line feed */
void text_crlf () {
  terminal_column = 0;
  terminal_row = (terminal_row < VGA_HEIGHT - 1) ? terminal_row + 1 : 0;
  if (terminal_row == 0) {
    terminal_clear();
  }
}

/* Writes a character to the current cursor position */
void terminal_putchar (char c) {
  const size_t index = terminal_row * VGA_WIDTH + terminal_column;

  switch (c) { /* ASCII control characters */
  case 0:
    break;
    
  case '\b':
    dec_wrap();
    terminal_putchar(' ');
    break;

  case '\f':
    terminal_clear();
    terminal_row = 0;
    terminal_column = 0;
    break;

  case '\n':
    text_crlf();
    break;

  case '\r':
    terminal_column = 0;
    break;
    
  default:
    terminal_buffer[index] = vga_entry(c, terminal_color);
    inc_wrap();
  }
}

/* Writes a fixed-length (portion of a) string */
void terminal_write_fixed (const char* data, size_t len) {
  size_t i;

  for (i = 0; i < len; i++) { 
    terminal_putchar(data[i]);
  }
}

/* Writes a string by writing each character */
void terminal_write (const char* data) {
  terminal_write_fixed(data, strlen(data));
}

/* Initializes the cursor and VGA buffer */
void terminal_init () {
  terminal_row = 0;
  terminal_column = 0;
  terminal_color = vga_entry_color(VGA_COLOR_LIGHT_GREY, VGA_COLOR_BLACK);
  terminal_buffer = (volatile uint16_t*) 0x000B8000;
}
